/*
 * This file is part of Kernel SPBM
 *
 * (c) 2003 Lehrstuhl Praktische Informatik IV, University of Mannheim
 *
 *   Kernel SPBM is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Kernel SPBM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <linux/mm.h>
#include <linux/ctype.h>
#include <linux/kernel.h>

#include "grid.h"
struct spbm_groups mygroups = { 0 };

uint16_t myseq = 0;

char *printGrid(struct grid *g);

struct dupentry {
	char addr[6];
	uint16_t seq;
};

LIST_HEAD(membertable);

struct updatepacket *ntoh_update(char *pkt, int len);
void getGroups(struct grid *g, struct spbm_groups *grouplist);
struct dupentry duptable[LOCDUPTABLE];
int dupptr = 0;

// true if packet is new
int checkDup(char *src, uint16_t seq)
{
	int i = 0;
	for (i = 0; i < LOCDUPTABLE; ++i) {
		if (!memcmp(duptable[i].addr, src, 6)) {
			if (seq == duptable[i].seq) {
				return 0;	// false
			}
		}
	}
	if (dupptr < (LOCDUPTABLE - 1))
		++dupptr;
	else
		dupptr = 0;
	memcpy(duptable[dupptr].addr, src, 6);
	duptable[dupptr].seq = seq;
	return 1;		// true 
}

char grouptmpstr[100];

char *printGroups(struct spbm_groups *groups)
{
	int i = 0;
	int len = 0;
	grouptmpstr[0] = 0;
	while (i < NUM_GROUPS) {
		if (GETGROUP(groups, i))
			len += sprintf(grouptmpstr + len, "%d ", i);
		i++;
	}
	return grouptmpstr;
}


int read_membertable(char *buffer, char **buffer_location, off_t offset,
		     int buffer_length, int *eof, void *data)
{
	int len = 0;
	struct list_head *n;
	len =
	    sprintf(buffer,
		    "targetgrid\tposition\tnode\t\t\tseq\tgroups\t\ttime\n");
	list_for_each(n, &membertable) {
		len +=
		    sprintf(buffer + len, "%s\t\t",
			    printGrid(&((struct member_entry *) n)->
				      targetgrid));
		if (((struct member_entry *) n)->haspos)
			len +=
			    sprintf(buffer + len, "%04x/%04x\t",
				    ((struct member_entry *) n)->position.
				    x,
				    ((struct member_entry *) n)->position.
				    y);
		else
			len += sprintf(buffer + len, "no pos     \t");
		len +=
		    sprintf(buffer + len, "%s\t",
			    addrToStr(((struct member_entry *) n)->node));
		len +=
		    sprintf(buffer + len, "%d\t",
			    ((struct member_entry *) n)->seq);
		len +=
		    sprintf(buffer + len, "%s\t",
			    printGroups(&((struct member_entry *) n)->
					groupvector));
		len +=
		    sprintf(buffer + len, "%d\n",
			    ((struct member_entry *) n)->timestamp);
	}
	return len;
}

void jitterSend(struct spbm_event *evt)
{
	sendPacket(evt->data, evt->data2, SPBM_TYPE_UPDATE, 0);
	kfree(evt->data);
	kfree(evt);
}


struct list_head *getMemberTable()
{
	return &membertable;
}

int updateEntry(struct grid *sourcegrid, struct spbm_groups *groups,
		char *addr, int seq, struct position *pos,
		struct timeval *t)
{
	struct member_entry *me = 0;
	struct member_entry *tmp;
	struct list_head *n;
	list_for_each(n, &membertable) {
		tmp = (struct member_entry *) n;
		if ((!gridcmp(&tmp->targetgrid, sourcegrid)) && tmp->haspos == (pos ? 1 : 0) && (((tmp->haspos == 0 || (tmp->haspos && !memcmp(addr, tmp->node, 6)))))) {	// if the grid is the same and if it is from the same node or not from any special node
			me = tmp;
		}
	}
	if (me) {
		if ((memcmp(addr, me->node, 6))
		    || (!memcmp(addr, me->node, 6)
			&& (seq > me->seq || seq + 200 < me->seq))) {
			memcpy(&(me->groupvector), groups,
			       sizeof(struct spbm_groups));
			if (pos) {
				memcpy(&(me->position), pos,
				       sizeof(struct position));
//      memcpy(&(me->targetgrid), sourcegrid, sizeof(struct grid)); // necessary?
				me->haspos = 1;
			} else
				me->haspos = 0;
			memcpy(&me->timestamp, t, sizeof(struct timeval));
			me->seq = seq;
			memcpy(me->node, addr, 6);
		} else
			return 0;
	} else {
		me = kmalloc(sizeof(struct member_entry), GFP_ATOMIC);
		list_add((struct list_head *) me, &membertable);
		memcpy(&(me->groupvector), groups,
		       sizeof(struct spbm_groups));
		memcpy(&(me->targetgrid), sourcegrid, sizeof(struct grid));
		if (pos) {
			memcpy(&(me->position), pos,
			       sizeof(struct position));
			me->haspos = 1;
		} else
			me->haspos = 0;
		memcpy(&me->timestamp, t, sizeof(struct timeval));
		me->seq = seq;
		memcpy(me->node, addr, 6);
	}
	return 1;
}

void processUpdate(struct spbmbeacon *spbm, char *addr, struct timeval *t,
		   int len)
{
	struct updatepacket *upd =
	    ntoh_update(((char *) spbm) + sizeof(struct spbmbeacon), len);
	struct grid sourcegrid;
	int deepestlevel = 0;
	int updated = 0;
	struct net_device *dev;

	struct grid mygrid;
	struct position mypos;

	// store nothing while position is unknown
	if (getXPos() == -1) {
		return;
	}

	mypos.x = getXPos();
	mypos.y = getYPos();

	posToGrid(&mypos, &mygrid);

	dev = getDev();
	if (!dev) {
		return;
	}
	if (!memcmp(upd->src, dev->dev_addr, 6)) {
		dev_put(dev);
		return;
	}
	dev_put(dev);

	if (!checkDup(upd->src, upd->seq)) {
		return;
	}
	posToGrid(&upd->pos, &sourcegrid);
//  printk("got update for %s (seq %d) from %s ",printGrid(&upd->destgrid),upd->seq,printGrid(&sourcegrid));
//  printk(" pos: %x/%x grid: \n",upd->pos.x,upd->pos.y);
	if (upd->destgrid.level == GRIDDEPTH)
		deepestlevel = 1;
	else
		sourcegrid.level -= GRIDDEPTH - upd->destgrid.level - 1;
	if (!gridcmp(&upd->destgrid, &mygrid)) {
		updated =
		    updateEntry(&sourcegrid, &upd->groups, upd->src,
				upd->seq, &upd->pos, t);
	} else if (deepestlevel) {
		updateEntry(&sourcegrid, &upd->groups, upd->src, upd->seq,
			    &upd->pos, t);
	} else if (storeInfo(&sourcegrid)) {
		updated =
		    updateEntry(&sourcegrid, &upd->groups, upd->src,
				upd->seq, 0, t);
	} else if (isSubGrid(&mygrid, &upd->destgrid)) {
		updated = 1;
	} else {
		// store without forwarding
		updateEntry(&sourcegrid, &upd->groups, upd->src, upd->seq,
			    0, t);
	}
	if (updated) {
		struct spbm_event *evt =
		    kmalloc(sizeof(struct spbm_event), GFP_ATOMIC);
		char *data =
		    kmalloc(len - sizeof(struct spbmbeacon), GFP_ATOMIC);
		memcpy(data, ((char *) spbm) + sizeof(struct spbmbeacon),
		       len - sizeof(struct spbmbeacon));
		event_set_time(evt, uniformint(10));
		evt->action = jitterSend;
		evt->data = data;
		evt->data2 = len - sizeof(struct spbmbeacon);
		event_add(evt);
	}
}

struct updatepacket tmpupdpkt;

// return value is only valid until next call to ntoh_update or hton_update
struct updatepacket *ntoh_update(char *pkt, int len)
{
	struct updatepacket *npkt = (struct updatepacket *) pkt;
	memcpy(&tmpupdpkt.destgrid, &npkt->destgrid, sizeof(struct grid));
	memcpy(tmpupdpkt.src, npkt->src, 6);

	tmpupdpkt.pos.x = ntohs(npkt->pos.x);
	tmpupdpkt.pos.y = ntohs(npkt->pos.y);

	memcpy(&tmpupdpkt.groups, &npkt->groups,
	       sizeof(struct spbm_groups));
	tmpupdpkt.seq = ntohs(npkt->seq);
	return &tmpupdpkt;
}

char *hton_update(struct updatepacket *pkt, int *len)
{
	struct updatepacket *npkt = (struct updatepacket *) pkt;
	memcpy(&tmpupdpkt.destgrid, &npkt->destgrid, sizeof(struct grid));
	memcpy(tmpupdpkt.src, npkt->src, 6);

	tmpupdpkt.pos.x = htons(npkt->pos.x);
	tmpupdpkt.pos.y = htons(npkt->pos.y);

	memcpy(&tmpupdpkt.groups, &npkt->groups,
	       sizeof(struct spbm_groups));
	tmpupdpkt.seq = htons(npkt->seq);
	*len = sizeof(struct updatepacket);
	return (char *) &tmpupdpkt;
}


void join(int group)
{
	SETGROUP(&mygroups, group);
}

void leave(int group)
{
	UNSETGROUP(&mygroups, group);
}

int isMember(int group)
{
	return GETGROUP(&mygroups, group);
}

// true if grid1 is subgrid of grid2
int isSubGrid(struct grid *grid1, struct grid *grid2)
{
	int i = 0;
	if (grid1->level < grid2->level)
		return 0;
	while (i < grid2->level) {
		if (grid1->grid[i] != grid2->grid[i])
			return 0;
		++i;
	}
	return 1;
}

int storeInfo(struct grid *g)
{
	int i = 0;
	struct grid mygrid;
	struct position mypos;

	// store nothing while position is unknown
	if (getXPos() == -1) {
		return 0;
	}

	mypos.x = getXPos();
	mypos.y = getYPos();

	posToGrid(&mypos, &mygrid);

	for (i = 0; i < g->level; ++i)
		if (mygrid.grid[i] != g->grid[i])
			break;
	if (i + 1 == g->level)
		return 1;
	return 0;
}

struct spbm_event updateevents[GRIDDEPTH + 1];

void sendUpdate(struct spbm_event *evt)
{
	struct sk_buff *skb = evt->data;
	int level = evt->data2;

	evt->data = 0;

	if (getXPos() == -1) {
		printk("Won't send update because position is unknown\n");
		goto out;
	}
	if (skb) {
//    dev_queue_xmit(skb);
	} else {
		struct updatepacket upkt;
		struct position mypos;
		int len;
		char *tosend;
		struct net_device *dev;
		upkt.pos.x = getXPos();
		upkt.pos.y = getYPos();

		upkt.seq = myseq++;

		mypos.x = getXPos();
		mypos.y = getYPos();
		posToGrid(&mypos, &upkt.destgrid);

		upkt.destgrid.level -= GRIDDEPTH - level;
		getGroups(&upkt.destgrid, &upkt.groups);
		upkt.destgrid.level--;
		dev = getDev();
		if (dev) {
			memcpy(&upkt.src, dev->dev_addr, dev->addr_len);
			dev_put(dev);
		}
		tosend = hton_update(&upkt, &len);
		sendPacket(tosend, len, SPBM_TYPE_UPDATE, 0);
	}

      out:
	event_set_time(evt,
		       (UPDATETIME +
			uniformint(UPDATEJITTER)) * (1 << (GRIDDEPTH -
							   (level - 1))));
	event_add(evt);
}

void initupdateevents()
{
	int i;
	for (i = 0; i <= GRIDDEPTH; ++i) {
		updateevents[i].action = sendUpdate;
		updateevents[i].data2 = i + 1;
		updateevents[i].data = 0;
		event_set_time(&updateevents[i],
			       (UPDATETIME +
				uniformint(UPDATEJITTER)) *
			       (1 << (GRIDDEPTH - i)));
		event_add(&updateevents[i]);
	}
	// dummy entries for debugging
#ifdef SINGLEDEBUG
	{
		struct grid tmpgrid;
		struct spbm_groups tmpgroups = { 0 };
		struct timeval t;
		tmpgrid.level = 3;
		tmpgrid.grid[0] = 0;
		tmpgrid.grid[1] = 0;
		tmpgrid.grid[2] = 2;
		SETGROUP(&tmpgroups, 1);
		updateEntry(&tmpgrid, &tmpgroups,
			    "\0xBA\0xBE\0xCA\0xFE\0xBA\0xBE", 1, 0, &t);
		initdummyneighbour();
	}
#endif
}

int gridcmp(struct grid *g1, struct grid *g2)
{
	int i = 0;
	if (g1->level != g2->level)
		return 1;
	while (i < g1->level) {
		if (g1->grid[i] != g2->grid[i])
			return 1;
		++i;
	}
	return 0;
}

// returns the groups for a grid. if g==0 for myself
void getGroupsPrivate(struct grid *g, struct spbm_groups *grouplist,
		      int withme)
{
	struct grid mygrid;
	struct position mypos;
	struct member_entry *gtmp;
	struct list_head *n;
	*((char *) grouplist) = 0;
	if (g == 0) {
		memcpy(grouplist, &mygroups, sizeof(struct spbm_groups));
		return;
	}

	mypos.x = getXPos();
	mypos.y = getYPos();
	posToGrid(&mypos, &mygrid);

	if (isSubGrid(&mygrid, g) || g->level > GRIDDEPTH) {
		// init with own groups
		if (withme)
			memcpy(grouplist, &mygroups,
			       sizeof(struct spbm_groups));
	} else {
		// init with zero
//    memset(grouplist, 0, sizeof(struct spbm_groups));
		gtmp = 0;
		list_for_each(n, &membertable) {
			if (!gridcmp
			    (&((struct member_entry *) n)->targetgrid,
			     g)) {
				gtmp = (struct member_entry *) n;
			}
		}
		if (gtmp) {	// found in membertable - return
			memcpy(grouplist, &gtmp->groupvector,
			       sizeof(struct spbm_groups));
			return;
		}
	}
	// construct groups from the things we know
	list_for_each(n, &membertable) {
		if (isSubGrid
		    (&(((struct member_entry *) n)->targetgrid), g)) {
			ORGROUPS(grouplist,
				 &(((struct member_entry *) n)->
				   groupvector));
		}
	}
	return;
}

void getGroups(struct grid *g, struct spbm_groups *grouplist)
{
	getGroupsPrivate(g, grouplist, 1);
}

void getGroupsNotMe(struct grid *g, struct spbm_groups *grouplist)
{
	getGroupsPrivate(g, grouplist, 0);
}

struct grid *posToGridLevel(struct position *pos, struct grid *g,
			    int level)
{
	int i = 0;
	int xtmp = 0;
	int ytmp = 0;
	g->grid[0] = 0;
	g->level = level;
	while (i < level) {
		xtmp = pos->x & (1 << (15 - i));
		ytmp = pos->y & (1 << (15 - i));
		g->grid[i] = (xtmp ? 2 : 0) + (ytmp ? 0 : 1);
		++i;
	}
	return g;
}

struct grid *posToGrid(struct position *pos, struct grid *g)
{
	return posToGridLevel(pos, g, GRIDDEPTH);
}

char gridname[GRIDDEPTH + 1];

char *printGrid(struct grid *g)
{
	int i = 0;
	int level = g->level;
	if (level > GRIDDEPTH)
		level = GRIDDEPTH;
	while (i < level) {
		gridname[i] = g->grid[i] + '1';
		++i;
	}
	gridname[i] = 0;
	return gridname;
}
