/*
 * This file is part of Kernel SPBM
 *
 * (c) 2003 Lehrstuhl Praktische Informatik IV, University of Mannheim
 *
 *   Kernel SPBM is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Kernel SPBM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include "module.h"
#include "thread.h"
#include <linux/netdevice.h>
#include <asm/byteorder.h>
#include "spbm.h"
#include <linux/if_ether.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/fs.h>
#include "neighbours.h"

char *dev;

MODULE_LICENSE("GPL");
MODULE_PARM(dev, "s");

int in_handler(struct sk_buff *, struct net_device *,
	       struct packet_type *);

//entries for our different PROC files
static struct proc_dir_entry *my_dir, *my_proc, *pos_file, *join_file,
    *leave_file, *member_file;

int xpos = -1;
int ypos = -1;

int getXPos()
{
	return xpos;
}

int getYPos()
{
	return ypos;
}

char addrbuf[24];

char *addrToStr(unsigned char *addr)
{
	sprintf(addrbuf, "%02x:%02x:%02x:%02x:%02x:%02x", addr[0], addr[1],
		addr[2], addr[3], addr[4], addr[5]);
	return addrbuf;
}

int read_pos(char *buffer, char **buffer_location, off_t offset,
	     int buffer_length, int *eof, void *data)
{
	int len;
	len = sprintf(buffer, "X: %d Y: %d", getXPos(), getYPos());
	return len;
}

extern struct spbm_event beaconevt;

int write_pos(struct file *file, const char *buffer, unsigned long count,
	      void *data)
{
	char temp[18];
	int xtmp, ytmp;
	int firstpos = (getXPos() == -1) ? 1 : 0;
	int len = (count < 17) ? count : 17;
	copy_from_user(temp, buffer, len);
	sscanf(temp, "%x %x", &xtmp, &ytmp);
	ypos = ytmp;
	xpos = xtmp;
	if (firstpos) {
		event_set_time(&beaconevt, 0);
		event_changed();
	}
	return len;
}

int read_join(char *buffer, char **buffer_location, off_t offset,
	      int buffer_length, int *eof, void *data)
{
	int len = 0;
	int i = 0;
	len = sprintf(buffer, "Member of: ");
	while (i < NUM_GROUPS) {
		if (isMember(i)) {
			len += sprintf(buffer + len, "%d ", i);
		}
		++i;
	}
	len += sprintf(buffer + len, "\n");
	return len;
}

int write_join(struct file *file, const char *buffer, unsigned long count,
	       void *data)
{
	char temp[18];
	int groupnum;
	int len = (count < 17) ? count : 17;
	copy_from_user(temp, buffer, len);
	sscanf(temp, "%x", &groupnum);
	join(groupnum);
	return len;
}

int write_leave(struct file *file, const char *buffer, unsigned long count,
		void *data)
{
	char temp[18];
	int groupnum;
	int len = (count < 17) ? count : 17;
	copy_from_user(temp, buffer, len);
	sscanf(temp, "%x", &groupnum);
	leave(groupnum);
	return len;
}

char mydev[100];
struct packet_type spbm_pt;
struct net_device *devi;

struct net_device *getDev()
{
	devi = (struct net_device *) dev_get_by_name(mydev);
	return devi;
}

void processOutPkt(struct spbm_event *evt)
{
	routeIPPacket((struct sk_buff *) evt->data);
	kfree_skb(evt->data);
	kfree(evt);
}

unsigned int packetgrabber(unsigned int hooknum, struct sk_buff **skb,
			   const struct net_device *in,
			   const struct net_device *out,
			   int (*okfn) (struct sk_buff *))
{
	struct spbm_event *evt;
	struct iphdr *ip = (*skb)->nh.iph;
//  printk("grabbed packet - ");
	if (!
	    ((ntohl(ip->daddr) < ntohl(GROUPOFFSET) + NUM_GROUPS)
	     && (ntohl(ip->daddr) >= ntohl(GROUPOFFSET)))) {
		printk("for different net (%x / %x)... passing\n",
		       ip->daddr, GROUPOFFSET);
		return NF_ACCEPT;
	}
	evt = kmalloc(sizeof(struct spbm_event), GFP_ATOMIC);
	printk("for me... stealing\n");
	evt->action = processOutPkt;
	evt->data = *skb;
	event_set_time(evt, 0);
	event_add(evt);

	return NF_STOLEN;
}

struct nf_hook_ops output_filter;

int read_membertable(char *buffer, char **buffer_location, off_t offset,
		     int buffer_length, int *eof, void *data);

int init_module(void)
{
	int result;

	printk("\nSPBM Kernel routing module - 2003 Thomas Butter ("
	       __DATE__ " " __TIME__ ")\n");
	fsync_dev(0);
	printk("synced all buffers\n");


	if (dev == NULL)
		strcpy(mydev, "eth0");
	else
		strcpy(mydev, dev);

	printk("using device: %s\n", mydev);

	spbm_pt.type = htons(PROT_SPBM);
	spbm_pt.dev = 0;
	spbm_pt.func = in_handler;
	spbm_pt.data = 0;


	startme();

	/* hooks registration */
	dev_add_pack(&spbm_pt);

	my_dir = proc_mkdir("spbm", NULL);
	my_dir->owner = THIS_MODULE;
	my_proc =
	    create_proc_read_entry("spbm/neighbours", 0, NULL, read_nb,
				   NULL);
	my_proc->owner = THIS_MODULE;

	member_file =
	    create_proc_read_entry("spbm/membertable", 0, NULL,
				   read_membertable, NULL);
	member_file->owner = THIS_MODULE;

	pos_file = create_proc_entry("pos", 0644, my_dir);

	pos_file->read_proc = read_pos;
	pos_file->write_proc = write_pos;
	pos_file->owner = THIS_MODULE;

	join_file = create_proc_entry("join", 0644, my_dir);

	join_file->read_proc = read_join;
	join_file->write_proc = write_join;
	join_file->owner = THIS_MODULE;

	leave_file = create_proc_entry("leave", 0644, my_dir);

	leave_file->read_proc = 0;
	leave_file->write_proc = write_leave;
	leave_file->owner = THIS_MODULE;

	// use netfilter to grab the packets before leaving the computer
	output_filter.hook = packetgrabber;
	output_filter.pf = PF_INET;
	output_filter.hooknum = NF_IP_LOCAL_OUT;

	nf_register_hook(&output_filter);

	printk("initialised\n");

	return 0;
}


void cleanup_module(void)
{
	// TODO free neighbourtable
	nf_unregister_hook(&output_filter);
	remove_proc_entry("neighbours", my_dir);
	remove_proc_entry("pos", my_dir);
	remove_proc_entry("membertable", my_dir);
	remove_proc_entry("spbm", NULL);
	dev_remove_pack(&spbm_pt);

	killme();
	printk("bye bye\n");
}

void processInPkt(struct spbm_event *evt)
{
	struct sk_buff *skb = evt->data;
	struct ethhdr *eth = skb->mac.ethernet;
	char *addr = eth->h_source;
	struct timeval *t = &(skb->stamp);
	struct spbmbeacon *spbm =
	    (struct spbmbeacon *) ((void *) eth + 14);
	int type = 0;

//    printk("processing packet"); 
	type = processBeacon(spbm, addr, t);
	if (type == 100) {
//          printk("packet ignored because sender is out of reach\n");
	}
	if (type == SPBM_TYPE_UPDATE) {
		processUpdate(spbm, addr, t, skb->len);
	}
	if (type == SPBM_TYPE_DATA) {
		routePacket(((char *) spbm) + sizeof(struct spbmbeacon),
			    skb->len - sizeof(struct spbmbeacon));
	}
	kfree_skb(skb);
	kfree(evt);
}

int in_handler(struct sk_buff *skb, struct net_device *nd,
	       struct packet_type *pt)
{
	struct net_device *dev = getDev();
	if (dev) {
		struct ethhdr *eth = skb->mac.ethernet;
		if (!memcmp(eth->h_dest, dev->dev_addr, 6) ||
		    !memcmp(eth->h_dest, dev->broadcast, 6)) {
			struct spbm_event *evt =
			    kmalloc(sizeof(struct spbm_event), GFP_ATOMIC);
			evt->action = processInPkt;
			evt->data = skb;
			event_set_time(evt, 0);
			event_add(evt);
		} else
			kfree_skb(skb);
		dev_put(dev);
	}
}
