/*
 * This file is part of Kernel SPBM
 *
 * (c) 2003 Lehrstuhl Praktische Informatik IV, University of Mannheim
 *
 *   Kernel SPBM is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Kernel SPBM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <linux/list.h>
#include <linux/mm.h>
#include "spbm.h"

LIST_HEAD(neighbours);

int comparetimeval(struct timeval tv1, struct timeval tv2)
{
	if (tv1.tv_sec != tv2.tv_sec)
		return tv1.tv_sec - tv2.tv_sec;
	else
		return tv1.tv_usec - tv2.tv_usec;
}

void cleanNeighbours(struct spbm_event *evt)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	struct list_head *n, *tmp;
	list_for_each_safe(n, tmp, &neighbours) {
		if (comparetimeval
		    (((struct spbmneighbour *) n)->timeout, tv) < 0) {
			list_del(n);
			kfree(n);
		}
	}
	event_set_time(evt, 500);
	event_add(evt);
}

int read_nb(char *buffer, char **buffer_location, off_t offset,
	    int buffer_length, int *eof, void *data)
{
	int len = 0;
	struct grid g;
	struct list_head *n;
	struct timeval tv;
	do_gettimeofday(&tv);
	len = sprintf(buffer, "now: %d\n", tv.tv_sec);
	list_for_each(n, &neighbours) {
		posToGrid(&(((struct spbmneighbour *) n)->pos), &g);
		len +=
		    sprintf(buffer + len, "%s %x %x %d ",
			    addrToStr(((struct spbmneighbour *) n)->addr),
			    ((struct spbmneighbour *) n)->pos.x,
			    ((struct spbmneighbour *) n)->pos.y,
			    ((struct spbmneighbour *) n)->timeout);
		len += sprintf(buffer + len, "%s\n", printGrid(&g));
	}
//      len=sprintf(buffer,"hallo");
	return len;
}

int processBeacon(struct spbmbeacon *spbm, char *addr, struct timeval *t)
{
	struct spbmneighbour *res = 0;
	struct list_head *n;
	int xpos = ntohs(spbm->pos.x);
	int ypos = ntohs(spbm->pos.y);
	long diffy = getYPos() - ypos;
	long diffx = getXPos() - xpos;

	if (getXPos() == -1)
		return 100;	// position unknown ignore everything
	if ((u32) diffy * (u32) diffy > (u32) MAXDISTSQUARE) {
		return 100;
	}
	if ((u32) (diffx * diffx) >
	    (unsigned long) MAXDISTSQUARE - (u32) diffy * diffy) {
//          printk("blah\n");
		return 100;
	}

	list_for_each(n, &neighbours) {
		if (memcmp(((struct spbmneighbour *) n)->addr, addr, 6) ==
		    0) {
			res = (struct spbmneighbour *) n;
		}
	}
	if (!res) {
		res =
		    (struct spbmneighbour *)
		    kmalloc(sizeof(struct spbmneighbour), GFP_ATOMIC);
		memcpy(res->addr, addr, 6);
		list_add((struct list_head *) res, &neighbours);
	}
	res->timeout.tv_sec = t->tv_sec;
	res->timeout.tv_usec = t->tv_usec + 2500 * BEACONTIME;
	while (res->timeout.tv_usec > 1000000) {
		res->timeout.tv_usec -= 1000000;
		res->timeout.tv_sec += 1;
	}
	res->pos.x = ntohs(spbm->pos.x);
	res->pos.y = ntohs(spbm->pos.y);
	return spbm->type;
}

#ifdef SINGLEDEBUG

struct spbmneighbour sn;

void initdummyneighbour()
{
	do_gettimeofday(&sn.timeout);
	sn.timeout.tv_sec += 20;
	sn.addr[0] = 11;
	sn.addr[1] = 11;
	sn.addr[2] = 11;
	sn.addr[3] = 11;
	sn.addr[4] = 11;
	sn.addr[5] = 11;
	sn.pos.x = 0x600;
	sn.pos.y = 0x500;
	list_add((struct list_head *) &sn, &neighbours);
}
#endif

u_int64_t distancesquare(struct position *p1, struct position *p2)
{
	return (u_int64_t) (p2->x - p1->x) * (u_int64_t) (p2->x - p1->x) +
	    (u_int64_t) (p2->y - p1->y) * (u_int64_t) (p2->y - p1->y);
}

char *findNearest(struct grid *g)
{
	struct list_head *n;
	struct position mypos;
	u_int64_t dist = (u_int64_t) 9999999999999.0;
	struct spbmneighbour *nearest = 0;
	u_int64_t mydist = 0;
	struct position gridmid = { 0, 0 };
	struct grid *tmpgrid;
	int i = 0;

	while (i < g->level) {
		if (g->grid[i] & 2) {
			gridmid.x += (1 << (15 - i));
		}
		if (!(g->grid[i] & 1)) {
			gridmid.y += (1 << (15 - i));
		}
		++i;
	}
	gridmid.x += (1 << (15 - i));
	gridmid.y += (1 << (15 - i));

	mypos.x = getXPos();
	mypos.y = getYPos();

	list_for_each(n, &neighbours) {
		posToGrid(&((struct spbmneighbour *) n)->pos, &tmpgrid);
		if (isSubGrid(&tmpgrid, g) &&
		    dist >
		    distancesquare(&((struct spbmneighbour *) n)->pos,
				   &gridmid)) {
			nearest = ((struct spbmneighbour *) n);
			dist =
			    distancesquare(&((struct spbmneighbour *) n)->
					   pos, &gridmid);
		}
	}
	mydist = distancesquare(&mypos, &gridmid);
	if (mydist < dist)
		dist = mydist;

	list_for_each(n, &neighbours) {
		if (dist >
		    distancesquare(&((struct spbmneighbour *) n)->pos,
				   &gridmid)) {
			dist =
			    distancesquare(&((struct spbmneighbour *) n)->
					   pos, &gridmid);
			nearest = ((struct spbmneighbour *) n);
		}
	}
	if (nearest)
		return nearest->addr;

	return 0;
}
