/*
 * This file is part of Kernel SPBM
 *
 * (c) 2003 Lehrstuhl Praktische Informatik IV, University of Mannheim
 *
 *   Kernel SPBM is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Kernel SPBM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/netdevice.h>
#include <linux/net.h>
#include <linux/udp.h>
#include <linux/inetdevice.h>

#include <linux/ip.h>
#include <net/checksum.h>
#include "grid.h"

void routePacket(char *pkt, int len);

char tmppkt[3000];		// MTU 1500 + enough header space

/**
 * routing header:
 * byte   group
 * byte   number of destination grids
 * struct grid * number of destination grids
 **/

void routeIPPacket(struct sk_buff *pkt)
{
	int groupnum;
	struct iphdr *ip = pkt->nh.iph;
	struct grid *destgrids = (struct grid *) &tmppkt[2];
	int i = 0;
	struct grid tmpgrid;
	struct grid mygrid;
	struct position pos;
	struct spbm_groups tmpgroups;
	int destcount = 0;

	groupnum = ntohl(ip->daddr) - ntohl(GROUPOFFSET);

	pos.x = getXPos();
	if (getXPos() == -1) {
		printk("position unknown - can't route");
		return;
	}
	pos.y = getYPos();
	posToGrid(&pos, &mygrid);

	tmppkt[0] = groupnum;	// groupnum in header for easier debugging

	// add the destination grids
	while (i < GRIDDEPTH) {
		char h = 0;
		tmpgrid.level = i + 1;
		while (h <= 3) {
			if (mygrid.grid[i] != h) {
				tmpgrid.grid[i] = h;
				getGroupsNotMe(&tmpgrid, &tmpgroups);
				if (GETGROUP(&tmpgroups, groupnum)) {
					memcpy(&destgrids[destcount++],
					       &tmpgrid,
					       sizeof(struct grid));
				}
			}
			++h;
		}
		tmpgrid.grid[i] = mygrid.grid[i];
		++i;
	}
	getGroupsNotMe(&tmpgrid, &tmpgroups);
	if (GETGROUP(&tmpgroups, groupnum)) {
		memcpy(&destgrids[destcount++], &tmpgrid,
		       sizeof(struct grid));
	}
	memcpy(&destgrids[destcount], ip, pkt->len);	// copy the ip packet after the header;
	tmppkt[1] = (char) destcount;
//  printk("destcount: %d\n",destcount);
	routePacket(tmppkt,
		    2 + sizeof(struct grid) * destcount + pkt->len);
}

char nexthops[30][6];
struct grid nextgrids[30][20];
int nextgridcount[30];
int nexthopcount;

int findNextHop(char *addr)
{
	if (!addr) {
		printk("addr==0!!!\n");
		return 0;
	}
	int i = 0;
	if (nexthopcount > 29) {
		printk("SPBM FATAL: too many nexthops\n");
		return 0;
	}
	while (i < nexthopcount) {
		if (!memcmp(nexthops[i], addr, 6)) {
			return i;
		}
		++i;
	}
	memcpy(nexthops[nexthopcount], addr, 6);
	nextgridcount[nexthopcount] = 0;
	nexthopcount++;
	return nexthopcount - 1;
}

char routePacketTmp[3000];

void routePacket(char *pkt, int len)
{
	int groupnum = pkt[0];
	int destcount = pkt[1];
	struct grid *dests = (struct grid *) &pkt[2];
	char *payload = pkt + 2 + sizeof(struct grid) * destcount;
	struct iphdr *iph = (struct iphdr *) payload;
	int i = 0;
	struct net_device *dev = getDev();
	struct grid mygrid;
	struct position pos;
	int pllen = len - 2 - sizeof(struct grid) * destcount;

	nexthopcount = 0;

	if (!dev) {
		printk("no dev - discarding packet");
		return;
	}

	iph->ttl--;

	if (iph->ttl == 0) {
		printk("time to live == 0 - discarding packet\n");
		return;
	}
	// TODO add csum check
	iph->check = 0;
	iph->check = ip_fast_csum((unsigned char *) iph, iph->ihl);

	if (isMember(groupnum)) {	// send packet to ip stack
		struct sk_buff *skb;
		struct net_device *dev = getDev();
		struct ethhdr *eth;
		struct in_device *indevi;
		struct iphdr *liph;
		char *pl;
		if (!dev)
			return;
		skb =
		    alloc_skb(len - (2 + sizeof(struct grid) * destcount) +
			      40, GFP_ATOMIC);
		skb_reserve(skb, 20);
		indevi = in_dev_get(dev);
		in_dev_put(indevi);
		if (dev)
			dev_put(dev);
		pl = skb_put(skb, pllen);
		memcpy(pl, payload, pllen);
		skb->dev = dev;
//    skb->dev=__dev_get_by_name("lo");
		liph = (struct iphdr *) skb->data;
		liph->daddr = indevi->ifa_list->ifa_address;
		skb->nh.iph = (struct iphdr *) pl;
		skb->h.raw = (unsigned char *) pl;
		liph->check = 0;
		liph->check =
		    ip_fast_csum((unsigned char *) liph, liph->ihl);
		if (liph->protocol == 17) {	// UDP - change checksum
			// 0 -> no checksum. RFC768 "An all zero  transmitted
			// checksum  value means that the transmitter  generated  no checksum  (for
			// debugging or for higher level protocols that don't care)."
			char *udp = ((char *) liph + (liph->ihl * 4));
			udp[6] = 0;
			udp[7] = 0;
			skb->h.raw = udp;
		}
		skb->mac.raw = (char *) liph;
		skb->dst = 0;

		skb->pkt_type = PACKET_HOST;	// HOST
		skb->ip_summed = 0;
		skb->protocol = htons(ETH_P_IP);


		netif_rx(skb);
	}

	pos.x = getXPos();
	if (getXPos() == -1) {
		printk("position unknown - can't route");
		return;
	}
	pos.y = getYPos();
	posToGrid(&pos, &mygrid);

	while (i < destcount) {
		if (isSubGrid(&mygrid, &dests[i])) {
			if (isSubGrid(&dests[i], &mygrid)) {	// <=> mygrid==dests[i]
				struct list_head *n;
				struct list_head *membertable =
				    getMemberTable();
				list_for_each(n, membertable) {
					struct member_entry *tmp =
					    (struct member_entry *) n;
					if (tmp->haspos
					    && isSubGrid(&tmp->targetgrid,
							 &mygrid)
					    && GETGROUP(&tmp->groupvector,
							groupnum)) {
						int num =
						    findNextHop(tmp->node);
					}
				}
			} else {	// find members in grids with prefix dests[i]
				int j = dests[i].level;
				struct grid tmpgrid;
				struct spbm_groups tmpgroups;
				memcpy(&tmpgrid, &dests[i],
				       sizeof(struct grid));
				while (j < GRIDDEPTH) {
					char h = 0;
					tmpgrid.grid[j] = 0;
					tmpgrid.level = j + 1;
					while (h <= 3) {
						if (mygrid.grid[j] != h) {
							tmpgrid.grid[j] =
							    h;
							getGroupsNotMe
							    (&tmpgrid,
							     &tmpgroups);
							if (GETGROUP
							    (&tmpgroups,
							     groupnum)) {
								char *nn =
								    findNearest
								    (&tmpgrid);
//              printk("a) dests[i] %s tmpgrid ",printGrid(&dests[i]));
//              printk("%s\n",printGrid(&tmpgrid));
								if (!nn) {
//                 printk("dropping grid %s because no greedy neighbour is known\n",printGrid(&tmpgrid));
								} else {
									int num = findNextHop(nn);
									memcpy
									    (&nextgrids
									     [num]
									     [nextgridcount
									      [num]++],
									     &tmpgrid,
									     sizeof
									     (struct
									      grid));
								}
							}
						}
						++h;
					}
					tmpgrid.grid[j] = mygrid.grid[j];
					++j;
				}
				getGroupsNotMe(&tmpgrid, &tmpgroups);
				if (GETGROUP(&tmpgroups, groupnum)) {	// adding my own grid...
					struct list_head *n;
					struct list_head *membertable =
					    getMemberTable();
					list_for_each(n, membertable) {
						struct member_entry *tmp =
						    (struct member_entry *)
						    n;
						if (tmp->haspos
						    && isSubGrid(&tmp->
								 targetgrid,
								 &mygrid)
						    && GETGROUP(&tmp->
								groupvector,
								groupnum))
						{
							int num =
							    findNextHop
							    (tmp->node);
						}
					}
//              char *nn=findNearest(&tmpgrid);
//              printk("dests[i] %s tmpgrid ",printGrid(&dests[i]));
//              printk("%s\n",printGrid(&tmpgrid));
//              if(!nn) {
//                 printk("dropping grid %s because no greedy neighbour is known\n",printGrid(&tmpgrid));
//              } else {
//                      int num=findNextHop(nn);
//                      memcpy(&nextgrids[nextgridcount[num]++],&tmpgrid,sizeof(struct grid));
//              }
				}
			}
		} else {
			char *nn = findNearest(&dests[i]);
//              printk("dests[i] %s\n ",printGrid(&dests[i]));
			if (!nn) {
//                 printk("dropping grid %s because no greedy neighbour is known\n",printGrid(&dests[i]));
			} else {
				int num = findNextHop(nn);
				memcpy(&nextgrids[num]
				       [nextgridcount[num]++], &dests[i],
				       sizeof(struct grid));
			}
		}
		++i;
	}
	i = 0;
	while (i < nexthopcount) {
		routePacketTmp[0] = groupnum;
		routePacketTmp[1] = nextgridcount[i];
		struct grid *g = (struct grid *) &routePacketTmp[2];
		char *pltmp;
		memcpy(g, nextgrids[i],
		       sizeof(struct grid) * nextgridcount[i]);
		pltmp =
		    ((char *) g) + nextgridcount[i] * sizeof(struct grid);
		memcpy(pltmp, payload, pllen);
		sendPacket(routePacketTmp,
			   pllen + 2 +
			   nextgridcount[i] * sizeof(struct grid),
			   SPBM_TYPE_DATA, nexthops[i]);
		++i;
	}
}
