/*
 * This file is part of Kernel SPBM
 *
 * (c) 2003 Lehrstuhl Praktische Informatik IV, University of Mannheim
 *
 *   Kernel SPBM is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Kernel SPBM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define DEVNAME "eth0"
//#include <linux/config.h>
//#include <linux/types.h>
#include <linux/list.h>
#include <linux/if_ether.h>
#include <linux/time.h>
//##include <netinet/in.h>
#include "config.h"

// http://www.cavebear.com/CaveBear/Ethernet/type.html
#define PROT_SPBM (unsigned short)(0x1705)

#define SPBM_TYPE_BEACON 1
#define SPBM_TYPE_UPDATE 2
#define SPBM_TYPE_DATA 3

struct grid {
	uint8_t level;
	uint8_t grid[GRIDDEPTH];
} __attribute__ ((packed));

struct position {
	uint16_t x;
	uint16_t y;
} __attribute__ ((packed));

struct spbmbeacon {
	uint8_t type;
	struct position pos;
} __attribute__ ((packed));

struct spbmneighbour {
	struct list_head list;
	unsigned char addr[ETH_ALEN];
	struct timeval timeout;
	struct position pos;
};

struct spbm_event {
	struct list_head list;
	struct timeval t;
	void (*action) (struct spbm_event *);
	void *data;
	int data2;
};

// for number of groups > 8 change the struct and the 4 defines below
#define NUM_GROUPS 8

struct spbm_groups {
	uint8_t groupset;
} __attribute__ ((packed));

#define SETGROUP(x,y)	(x)->groupset|=(1<<(y));
#define UNSETGROUP(x,y)	(x)->groupset= ( (0xff-(1<<(y))) & ((x)->groupset));
#define GETGROUP(x,y)	(x)->groupset & (1<<(y))
#define ORGROUPS(x,y)	(x)->groupset|=(y)->groupset

struct member_entry {
	struct list_head list;
	struct grid targetgrid;
	struct position position;
	char node[6];
	int seq;
	struct spbm_groups groupvector;
	struct timeval timestamp;
	int haspos;
};

struct net_device *getDev();
char *addrToStr(unsigned char *addr);
void sendPacket(char *payload, int len, int type, char *dest);
